package billing

import (
	"context"
	"crypto/hmac"
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"hash"
	"strconv"
	"strings"
	"time"

	"github.com/google/uuid"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
)

var log = _log.New("billing")

var (
	confEnabled            = false
	confRegionName         = ""
	confInfraName          = ""
	confInfraType          = ""
	confResellerPrefix     = "AUTH_"
	confStoragePolicyName  = "OIO"
	confStoragePolicyIndex = 2
	confBatchSize          = 50
	confSecret             = "SECRET"

	confInterval = 15 * time.Minute

	confAMQPURLs       = "amqp://guest:guest@localhost:5672/"
	confAMQPExchange   = ""
	confAMQPKey        = "telemetry"
	confAMQPPersistent = false
	confAMQPPriority   uint

	confAMQPCACertFile   = ""
	confAMQPCertFile     = ""
	confAMQPKeyFile      = ""
	confAMQPDepth        uint
	confAMQPFailIfNoPeer = true
	confAMQPVerify       = true
)

func InitFlags(fs *flag.FlagSet) {
	fs.BoolVar(&confEnabled, "billing.enable", confEnabled, "[billing] enable")
	fs.StringVar(&confRegionName, "billing.region_name", confRegionName, "[billing] region name")
	fs.StringVar(&confInfraName, "billing.infra_name", confInfraName, "[billing] infra name")
	fs.StringVar(&confInfraType, "billing.infra_type", confInfraType, "[billing] infra type")
	fs.StringVar(&confResellerPrefix, "billing.reseller_prefix", confResellerPrefix, "[billing] reseller prefix")
	fs.StringVar(&confStoragePolicyName, "billing.storage_policy_name", confStoragePolicyName, "[billing] storage policy name")
	fs.IntVar(&confStoragePolicyIndex, "billing.storage_policy_index", confStoragePolicyIndex, "[billing] storage policy index")
	fs.IntVar(&confBatchSize, "billing.batch_size", confBatchSize, "[billing] samples batch size")
	fs.StringVar(&confSecret, "billing.secret", confSecret, "[billing] signing secret")
	fs.StringVar(&confAMQPURLs, "billing.amqp_urls", confAMQPURLs, "[billing] amqp URLs (comma-separated list)")
	fs.StringVar(&confAMQPExchange, "billing.amqp_exchange", confAMQPExchange, "[billing] amqp exchange")
	fs.StringVar(&confAMQPKey, "billing.amqp_key", confAMQPKey, "[billing] amqp key")
	fs.BoolVar(&confAMQPPersistent, "billing.amqp_persistent", confAMQPPersistent, "[billing] amqp messages persistency")
	fs.UintVar(&confAMQPPriority, "billing.amqp_priority", confAMQPPriority, "[billing] amqp messages priority (0-9)")
	fs.StringVar(&confAMQPCACertFile, "billing.amqp_cacertfile", confAMQPCACertFile, "[billing] amqp SSL CA certificate file")
	fs.StringVar(&confAMQPCertFile, "billing.amqp_certfile", confAMQPCertFile, "[billing] amqp SSL certificate file")
	fs.StringVar(&confAMQPKeyFile, "billing.amqp_keyfile", confAMQPKeyFile, "[billing] amqp SSL key file")
	fs.UintVar(&confAMQPDepth, "billing.amqp_depth", confAMQPDepth, "[billing] amqp SSL verification depth (see RabbitMQ doc)")
	fs.BoolVar(&confAMQPFailIfNoPeer, "billing.amqp_failifnopeer", confAMQPFailIfNoPeer, "[billing] amqp SSL fail if no peer (see RabbitMQ doc)")
	fs.BoolVar(&confAMQPVerify, "billing.amqp_verify", confAMQPVerify, "[billing] amqp SSL verify (see RabbitMQ doc)")
}

func Enabled() bool {
	return confEnabled || (confRegionName != "") || (confInfraName != "") || (confInfraType != "")
}

type BillingDaemon struct {
	metrics  *common.MetricData
	interval time.Duration

	regionName         string
	infraName          string
	infraType          string
	resellerPrefix     string
	storagePolicyName  string
	storagePolicyIndex int
	batchSize          int
	digestMaker        hash.Hash

	amqpClient *amqpClient
}

func NewBillingDaemon(m *common.MetricData) *BillingDaemon {
	if confRegionName == "" {
		log.Fatal("Billing was enabled, but region_name was not set")
	}
	if confInfraName == "" {
		log.Fatal("Billing was enabled, but infra_name was not set")
	}
	if confInfraType == "" {
		log.Fatal("Billing was enabled, but infra_type was not set")
	}

	clientConfig := &amqpClientConfig{
		urls:         confAMQPURLs,
		exchange:     confAMQPExchange,
		key:          confAMQPKey,
		persistent:   confAMQPPersistent,
		priority:     confAMQPPriority,
		caCertFile:   confAMQPCACertFile,
		certFile:     confAMQPCertFile,
		keyFile:      confAMQPKeyFile,
		depth:        confAMQPDepth,
		failIfNoPeer: confAMQPFailIfNoPeer,
		verify:       confAMQPVerify,
	}
	amqpClient, err := newAMQPClient(clientConfig)
	if err != nil {
		log.Fatalf("Could not create AMQP client: %s", err)
	}

	return &BillingDaemon{
		interval:           confInterval,
		metrics:            m,
		regionName:         confRegionName,
		infraName:          confInfraName,
		infraType:          confInfraType,
		resellerPrefix:     confResellerPrefix,
		storagePolicyName:  confStoragePolicyName,
		storagePolicyIndex: confStoragePolicyIndex,
		batchSize:          confBatchSize,
		digestMaker:        hmac.New(sha256.New, []byte(confSecret)),
		amqpClient:         amqpClient,
	}
}

func (d *BillingDaemon) Run(ctx context.Context) {
	ticker := time.NewTicker(d.interval)
	defer ticker.Stop()

	tickerC := ticker.C
	doneC := ctx.Done()

	go d.amqpClient.Start(ctx)

loop:
	for {
		if err := d.getAndPushBillingMetrics(ctx); err != nil {
			log.Errorf("Billing error: %s", err)
			// TODO: retry sooner?
		}

		select {
		case <-doneC:
			log.Warn("Close billing")

			break loop
		case <-tickerC:
		}
	}

	if err := d.Close(); err != nil {
		log.Warn("Closing billing daemon: %s", err)
	}
}

func (d *BillingDaemon) Close() error {
	if err := d.amqpClient.Close(); err != nil {
		return fmt.Errorf("Closing AMQP client: %w", err)
	}

	return nil
}

func (d *BillingDaemon) getAndPushBillingMetrics(ctx context.Context) error {
	accounts := d.getAccountMetrics()
	if len(accounts) == 0 {
		return errors.New("No account")
	}

	samples, err := d.makeBillingSamples(accounts)
	if err != nil {
		return fmt.Errorf("Making samples: %w", err)
	}

	nbSamples := len(samples)

	batchSize := d.batchSize
	if d.batchSize == 0 {
		batchSize = nbSamples + 1
	}

	for i := 0; i < nbSamples; i += batchSize {
		iEnd := i + batchSize
		if iEnd > nbSamples {
			iEnd = nbSamples
		}

		samplesBatch := samples[i:iEnd]
		payload, err := d.makeBillingPayload(samplesBatch)
		if err != nil {
			return fmt.Errorf("Making payload: %w", err)
		}

		if err := d.pushBillingMessage(ctx, payload); err != nil {
			return fmt.Errorf("Pushing message: %w", err)
		}
	}

	return nil
}

type accountInfo struct {
	Objects, Containers, Size float64
	PollingTimestamp          time.Time
}

func (d *BillingDaemon) getAccountMetrics() map[string]*accountInfo {
	metrics := d.metrics
	accounts := make(map[string]*accountInfo)

	metrics.RLock()
	for _, metric := range metrics.Measures {
		metricName := metric.Name

		if !strings.HasPrefix(metricName, "account_") {
			continue
		}

		metricType := metricName[len("account_"):]
		if metricType == "buckets" {
			continue
		}

		metricAccount, _ := metric.GetLabel("account")
		if _, ok := accounts[metricAccount]; !ok {
			accounts[metricAccount] = &accountInfo{}
		}

		metricValue := metric.Value

		switch metricType {
		case "objects":
			accounts[metricAccount].Objects = metricValue
		case "containers":
			accounts[metricAccount].Containers = metricValue
		case "size_kibibytes":
			accounts[metricAccount].Size = metricValue
			accounts[metricAccount].PollingTimestamp = metric.GetTimestamp()
		}
	}
	metrics.RUnlock()

	return accounts
}

func (d *BillingDaemon) makeBillingSamples(accounts map[string]*accountInfo) ([]*billingSample, error) {
	resellerPrefix := d.resellerPrefix
	digestMaker := d.digestMaker

	samples := make([]*billingSample, 0, len(accounts))
	for accountName, account := range accounts {
		if !strings.HasPrefix(accountName, resellerPrefix) {
			log.Debugf(`Account "%s" does not start with reseller prefix "%s", skipping`, accountName, resellerPrefix)
			continue
		}

		accountSize := int64(account.Size) * 1024
		accountContainers := int64(account.Containers)
		accountObjects := int64(account.Objects)
		tsISO8601 := account.PollingTimestamp.UTC().Format("2006-01-02T15:04:05Z")
		projectID := accountName[len(resellerPrefix):]

		// UUID v1
		messageID, err := uuid.NewUUID()
		if err != nil {
			return nil, fmt.Errorf("Getting an UUID v1: %w", err)
		}

		sample := &billingSample{
			CounterName:   "storage.account.objects.size",
			CounterType:   "gauge",
			CounterUnit:   "B",
			CounterVolume: accountSize,
			MessageID:     messageID.String(),
			ProjectID:     projectID,
			ResourceID:    projectID,
			ResourceMetadata: &resourceMetadata{
				InfraName: d.infraName,
				InfraType: d.infraType,
				PolicyStat: []*policyStat{
					{
						BytesUsed:          accountSize,
						ContainerCount:     accountContainers,
						ObjectCount:        accountObjects,
						StoragePolicyIndex: d.storagePolicyIndex,
						StoragePolicyName:  d.storagePolicyName,
					},
				},
				RegionName: d.regionName,
			},
			Source:    d.regionName,
			Timestamp: tsISO8601, // should be one of the metrics timestamp, but which one?
		}

		sample.sign(digestMaker)

		samples = append(samples, sample)
	}

	return samples, nil
}

func (d *BillingDaemon) makeBillingPayload(samples []*billingSample) (*messagePayload, error) {
	now := time.Now().UTC().Format("2006-01-02 15:04:05.999999")

	// UUID v4
	messageID, err := uuid.NewRandom()
	if err != nil {
		return nil, fmt.Errorf("Getting an UUID v4: %w", err)
	}
	messageIDStr := messageID.String()

	// UUID v4
	uniqueID, err := uuid.NewRandom()
	if err != nil {
		return nil, fmt.Errorf("Getting an UUID v4: %w", err)
	}
	uniqueIDStr := strings.Replace(uniqueID.String(), "-", "", 4)

	// https://github.com/openstack/oslo.messaging/blob/e44c9883066d9b2d081a594b97aac3d598d491c9/oslo_messaging/notify/notifier.py#L296
	// Payload: https://github.com/openstack/ceilometer/blob/e3f4be648747b8f46961a7ae977dffb1808ec755/ceilometer/polling/manager.py#L231
	// UniqueID: https://github.com/openstack/oslo.messaging/blob/e44c9883066d9b2d081a594b97aac3d598d491c9/oslo_messaging/_drivers/amqp.py#L121
	payload := &messagePayload{
		EventType:   "telemetry.polling",
		MessageID:   messageIDStr,
		Priority:    "SAMPLE",
		PublisherID: "ceilometer.polling",
		Timestamp:   now,
		UniqueID:    uniqueIDStr,
		Payload: &payloadSamples{
			Samples: samples,
		},
	}

	return payload, nil
}

func (d *BillingDaemon) pushBillingMessage(ctx context.Context, payload *messagePayload) error {
	marshalledPayload, err := json.Marshal(payload)
	if err != nil {
		return fmt.Errorf("Marshalling the payload: %w", err)
	}

	message := &osloMessage{
		OsloMessage: string(marshalledPayload),
		OsloVersion: "2.0",
	}
	messageID := payload.MessageID

	bytes, err := json.Marshal(message)
	if err != nil {
		return fmt.Errorf("Marshalling message: %w", err)
	}

	if err := d.amqpClient.Push(ctx, messageID, bytes); err != nil {
		return fmt.Errorf("Pushing message: %w", err)
	}

	log.Info("Billing message published to the AMQP queue")

	return nil
}

type osloMessage struct {
	OsloMessage string `json:"oslo.message"`
	OsloVersion string `json:"oslo.version"`
}

type messagePayload struct {
	ContextAuthToken      *string  `json:"_context_auth_token"`
	ContextDomain         *string  `json:"_context_domain"`
	ContextIsAdmin        *bool    `json:"_context_is_admin"`
	ContextIsAdminProject *bool    `json:"_context_is_admin_project"`
	ContextProjectDomain  *string  `json:"_context_project_domain"`
	ContextReadOnly       *bool    `json:"_context_read_only"`
	ContextRequestID      *string  `json:"_context_request_id"`
	ContextResourceUUID   *string  `json:"_context_resource_uuid"`
	ContextRoles          []string `json:"_context_roles"`
	ContextShowDeleted    *bool    `json:"_context_show_deleted"`
	ContextTenant         *string  `json:"_context_tenant"`
	ContextUser           *string  `json:"_context_user"`
	ContextUserDomain     *string  `json:"_context_user_domain"`
	ContextUserIdentity   *string  `json:"_context_user_identity"`
	EventType             string   `json:"event_type"`
	MessageID             string   `json:"message_id"`
	Priority              string   `json:"priority"`
	PublisherID           string   `json:"publisher_id"`
	Timestamp             string   `json:"timestamp"`
	UniqueID              string   `json:"_unique_id"`

	Payload *payloadSamples `json:"payload"`
}

type payloadSamples struct {
	Samples []*billingSample `json:"samples"`
}

type billingSample struct {
	CounterName      string            `json:"counter_name"`
	CounterType      string            `json:"counter_type"`
	CounterUnit      string            `json:"counter_unit"`
	CounterVolume    int64             `json:"counter_volume"`
	MessageID        string            `json:"message_id"`
	MessageSignature string            `json:"message_signature"`
	ProjectID        string            `json:"project_id"`
	ResourceID       string            `json:"resource_id"`
	ResourceMetadata *resourceMetadata `json:"resource_metadata"`
	Source           string            `json:"source"`
	Timestamp        string            `json:"timestamp"`
	UserID           *string           `json:"user_id"`
}

// following
// https://github.com/openstack/ceilometer/blob/4ae919c96e4116ab83e5d83f2b726ed44d165278/ceilometer/publisher/utils.py#L78
// a bit sketchy, but it shouldn't be used anymore in the future
func (s *billingSample) sign(digestMaker hash.Hash) {
	digestMaker.Reset()

	// result of:
	// https://github.com/openstack/ceilometer/blob/4ae919c96e4116ab83e5d83f2b726ed44d165278/ceilometer/publisher/utils.py#L42
	policyStatTpl := `{b'bytes_used': %d, b'container_count': %d, b'object_count': %d, b'storage_policy_index': %d, b'storage_policy_name': b'%s'}`

	_, _ = digestMaker.Write([]byte("counter_name"))
	_, _ = digestMaker.Write([]byte(s.CounterName))
	_, _ = digestMaker.Write([]byte("counter_type"))
	_, _ = digestMaker.Write([]byte(s.CounterType))
	_, _ = digestMaker.Write([]byte("counter_unit"))
	_, _ = digestMaker.Write([]byte(s.CounterUnit))
	_, _ = digestMaker.Write([]byte("counter_volume"))
	_, _ = digestMaker.Write([]byte(strconv.FormatInt(s.CounterVolume, 10)))
	_, _ = digestMaker.Write([]byte("message_id"))
	_, _ = digestMaker.Write([]byte(s.MessageID))
	_, _ = digestMaker.Write([]byte("monotonic_time"))
	_, _ = digestMaker.Write([]byte("None"))
	_, _ = digestMaker.Write([]byte("project_id"))
	_, _ = digestMaker.Write([]byte(s.ProjectID))
	_, _ = digestMaker.Write([]byte("resource_id"))
	_, _ = digestMaker.Write([]byte(s.ResourceID))
	_, _ = digestMaker.Write([]byte("resource_metadata:infra_name"))
	_, _ = digestMaker.Write([]byte(s.ResourceMetadata.InfraName))
	_, _ = digestMaker.Write([]byte("resource_metadata:infra_type"))
	_, _ = digestMaker.Write([]byte(s.ResourceMetadata.InfraType))
	_, _ = digestMaker.Write([]byte("resource_metadata:policy_stat"))

	nbPolicies := len(s.ResourceMetadata.PolicyStat)
	_, _ = digestMaker.Write([]byte("["))
	for i, policyStat := range s.ResourceMetadata.PolicyStat {
		fmt.Fprintf(digestMaker, policyStatTpl,
			policyStat.BytesUsed,
			policyStat.ContainerCount,
			policyStat.ObjectCount,
			policyStat.StoragePolicyIndex,
			policyStat.StoragePolicyName)

		if i < nbPolicies-1 {
			_, _ = digestMaker.Write([]byte(", "))
		}
	}
	_, _ = digestMaker.Write([]byte("]"))

	_, _ = digestMaker.Write([]byte("resource_metadata:region_name"))
	_, _ = digestMaker.Write([]byte(s.ResourceMetadata.RegionName))
	_, _ = digestMaker.Write([]byte("source"))
	_, _ = digestMaker.Write([]byte(s.Source))
	_, _ = digestMaker.Write([]byte("timestamp"))
	_, _ = digestMaker.Write([]byte(s.Timestamp))
	_, _ = digestMaker.Write([]byte("user_id"))
	_, _ = digestMaker.Write([]byte("None"))

	digest := hex.EncodeToString(digestMaker.Sum(nil))

	s.MessageSignature = digest
}

type resourceMetadata struct {
	InfraName  string        `json:"infra_name"`
	InfraType  string        `json:"infra_type"`
	PolicyStat []*policyStat `json:"policy_stat"`
	RegionName string        `json:"region_name"`
}

type policyStat struct {
	BytesUsed          int64  `json:"bytes_used"`
	ContainerCount     int64  `json:"container_count"`
	ObjectCount        int64  `json:"object_count"`
	StoragePolicyIndex int    `json:"storage_policy_index"`
	StoragePolicyName  string `json:"storage_policy_name"`
}
