package version

import (
	"context"
	"encoding/base64"
	"errors"
	"flag"
	"fmt"
	"os"
	"os/exec"
	"regexp"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
	"gopkg.in/yaml.v2"
)

var log = _log.New("version")

var confTimeout = 5 * time.Second
var confInterval = 5 * time.Minute

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "version.timeout", confTimeout, "[version] command timeout")
	fs.DurationVar(&confInterval, "version.interval", confInterval, "[version] interval between polls")
}

type versionDaemonFactory struct {
	defaultVersions string
	versionsPath    string

	packageCmd string
	packager   string

	versions Versions
	version  int
}

func NewFactory(defaultVersions, versionsPath string) daemon.DaemonFactoryI {
	return &versionDaemonFactory{
		defaultVersions: defaultVersions,
		versionsPath:    versionsPath,
	}
}

func (df *versionDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	versions := df.versions
	services := inventory.Services
	names := make(map[string]struct{})
	for _, version := range versions {
		serviceType := version.Service
		_, ok := services[serviceType]
		if serviceType != "" && !ok {
			continue
		}

		names[version.Name] = struct{}{}
	}

	return map[string]interface{}{
		"version": &versionDaemonParams{
			Namespace: inventory.Namespace,
			Version:   df.version,
			Names:     names,
		},
	}
}

// TODO: don't do anything if the file hasn't changed
func (df *versionDaemonFactory) Reload() error {
	defaultVersions := df.defaultVersions
	versionsPath := df.versionsPath

	var packageCmd string
	var packager string
	if checkCMD("rpm") {
		packageCmd = "rpm -q --qf %{VERSION}-%{RELEASE}"
		packager = "rpm"
	} else if checkCMD("dpkg-query") {
		packageCmd = "dpkg-query -Wf=${Version}"
		packager = "deb"
	} else {
		return errors.New("Failed to determine distro")
	}

	var versions struct {
		Versions Versions
	}

	if versionsPath == "" {
		data, err := base64.StdEncoding.DecodeString(defaultVersions)
		if err != nil {
			return err
		}

		if err := yaml.Unmarshal(data, &versions); err != nil {
			return err
		}
		log.Info("versions definition loaded from defaults")
	} else {
		f, err := os.Open(versionsPath)
		if err != nil {
			return err
		}
		defer f.Close()

		if err := yaml.NewDecoder(f).Decode(&versions); err != nil {
			return err
		}
		log.Infof("versions definition loaded from %s", versionsPath)
	}

	for _, version := range versions.Versions {
		if version.Name == "" {
			return errors.New("name is missing in one of the version definition")
		}

		switch version.Type {
		case "package":
			if version.Package == "" && (version.Deb == "" || version.Rpm == "") {
				return fmt.Errorf("package type must have a package key or rpm and deb keys: %s", version.Name)
			}
		case "command":
			if version.Command == "" {
				return fmt.Errorf("command type must have a command: %s", version.Name)
			}
			if version.Extract != "" {
				re, err := regexp.Compile(version.Extract)
				if err != nil {
					return fmt.Errorf("invalid regex for %s: %w", version.Name, err)
				}
				version.Regex = re
			}
		default:
			return fmt.Errorf("unknown type %s for %s", version.Type, version.Name)
		}
	}

	if !versions.Versions.Equal(df.versions) {
		log.Infoln("Versions updated")

		df.version++
	}

	df.packageCmd = packageCmd
	df.packager = packager
	df.versions = versions.Versions

	log.Infof("loaded %d versions definitions", len(df.versions))

	return nil
}

func checkCMD(cmd string) bool {
	_, err := exec.LookPath(cmd)
	return err == nil
}

func (df *versionDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*versionDaemonParams)
	versions := df.versions
	names := params.Names

	filteredVersions := make(Versions, 0, len(versions))
	for _, version := range versions {
		if _, ok := names[version.Name]; !ok {
			continue
		}

		filteredVersions = append(filteredVersions, version)
	}

	return &versionDaemon{
		packageCmd: df.packageCmd,
		packager:   df.packager,
		versions:   filteredVersions,
		namespace:  params.Namespace,
	}
}

type Versions []*Version

type Version struct {
	Name    string         `yaml:"name"`
	Type    string         `yaml:"type"`
	Package string         `yaml:"package"`
	Rpm     string         `yaml:"rpm"`
	Deb     string         `yaml:"deb"`
	Command string         `yaml:"command"`
	Extract string         `yaml:"extract"`
	Service string         `yaml:"service"`
	Regex   *regexp.Regexp `yaml:"_disabled2"`
}

func (versions1 Versions) Equal(versions2 Versions) bool {
	if len(versions1) != len(versions2) {
		return false
	}

	for _, v1 := range versions1 {
		found := false
		for _, v2 := range versions2 {
			if v1.Name != v2.Name {
				continue
			}

			if !v1.Equal(v2) {
				return false
			}

			found = true
		}

		if !found {
			return false
		}
	}

	return true
}

// To exclude Regex
func (v1 *Version) Equal(v2 *Version) bool {
	return v2 != nil &&
		v1.Name == v2.Name &&
		v1.Type == v2.Type &&
		v1.Package == v2.Package &&
		v1.Rpm == v2.Rpm &&
		v1.Deb == v2.Deb &&
		v1.Command == v2.Command &&
		v1.Extract == v2.Extract
}

type versionDaemonParams struct {
	Namespace string
	Version   int
	Names     map[string]struct{}
}

type versionDaemon struct {
	daemon.BaseDaemon

	namespace string

	packageCmd string
	packager   string

	versions []*Version
}

func (d *versionDaemon) Interval() time.Duration {
	return confInterval
}

func (d *versionDaemon) Delay() time.Duration {
	return 0
}

func (d *versionDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	for _, version := range d.versions {
		var cmd string
		var regex *regexp.Regexp

		var pkg string
		if version.Type == "package" {
			cmd = d.packageCmd + " "
			if version.Package != "" {
				pkg = version.Package
			} else if d.packager == "rpm" && version.Rpm != "" {
				pkg = version.Rpm
			} else if d.packager == "deb" && version.Deb != "" {
				pkg = version.Deb
			} else {
				continue
			}
			cmd += pkg
		} else if version.Type == "command" {
			cmd = version.Command
			regex = version.Regex
			pkg = version.Name
		}

		out, err := d.runCommand(ctx, cmd, regex)
		// TODO: do something if err == exec.ErrNotFound
		if err != nil {
			m.IncrError("command", version.Name)
			log.Warnln("error while running command", cmd, version.Name, err)
			continue
		}

		m.Set(&common.Metric{
			Name: "version",
			Labels: map[string]string{
				"namespace": d.namespace,
				"software":  version.Name,
				"package":   pkg,
				"version":   out,
			},
			Value: float64(time.Now().Unix()),
			TTL:   confInterval * 2,
		})
	}

	return nil
}

func (d *versionDaemon) runCommand(ctx context.Context, cmd string, regex *regexp.Regexp) (string, error) {
	out, err := common.Exec(ctx, confTimeout, cmd)
	if err != nil {
		return "", err
	}

	outFmt := strings.Trim(strings.TrimSuffix(string(out), "\n"), " ")

	if regex == nil {
		return outFmt, nil
	}

	matches := regex.FindStringSubmatch(outFmt)
	if len(matches) < 1 {
		return "", nil
	}
	if len(matches) == 1 {
		return matches[0], nil
	}

	return matches[1], nil
}
